use std::{
    env,
    time::{Duration, Instant},
};

use log::{debug, info, warn};
use serial::{CharSize, Parity, SerialPort, StopBits};

const PATTERN_COUNT: usize = 1;
const PATTERN: [[u8; 19]; PATTERN_COUNT] = [[
    0x77, 0x07, 0x01, 0x00, 0x01, 0x08, 0x00, 0xFF, 0x64, 0x00, 0x01, 0x82, 0x01, 0x62, 0x1E, 0x52,
    0xFF, 0x56, 0x00,
]];

#[tokio::main]
async fn main() {
    dotenvy::dotenv().ok();
    env_logger::init();
    loop {
        let mut port =
            serial::open(&env::var("USB_PORT").expect("USB_PORT environment variable")).unwrap();

        interact(&mut port).await.unwrap();
    }
}

async fn interact<T: SerialPort>(port: &mut T) -> anyhow::Result<()> {
    port.reconfigure(&|settings| {
        settings.set_baud_rate(serial::Baud9600).unwrap();
        settings.set_parity(Parity::ParityNone);
        settings.set_char_size(CharSize::Bits8);
        settings.set_stop_bits(StopBits::Stop1);
        Ok(())
    })
    .unwrap();

    port.set_timeout(Duration::from_millis(250)).unwrap();

    let mut read_buf: [u8; 1024] = [0; 1024];
    let mut queue: Vec<u8> = vec![];

    let mut prev_value = 0;
    let mut prev_time = Instant::now();

    let mut brake = 0_u32;
    let brake_target: u32 = 1;

    let mut empty_messages = 0;
    loop {
        let read_result = port.read(&mut read_buf);
        if let Ok(queue_length) = read_result {
            empty_messages = 0;
            queue.append(
                &mut read_buf
                    .iter()
                    .take(queue_length)
                    .copied()
                    .collect::<Vec<u8>>(),
            );
        } else if !queue.is_empty() && brake >= brake_target {
            let result = extract_values(&mut queue);

            if prev_value > 0 {
                let val =
                    ((result[0] - prev_value) * 3600) as f64 / (prev_time.elapsed().as_secs_f64());
                if !(0.0..=200000.0).contains(&val) || result[0] > 1000000000 {
                    warn!("received invalid data, dropping it");
                    continue;
                }

                info!(
                    "Zählerstand: {} Wh, Aktueller Verbrauch: {} W",
                    result[0] / 10,
                    val / 10.0
                );
            }

            prev_value = result[0];
            prev_time = Instant::now();
            brake = 0;
        } else if !queue.is_empty() {
            queue.clear();
            brake += 1;
        } else if empty_messages >= 40 {
            return match read_result {
                Ok(_) => Ok(()),
                Err(err) => Err(err.into()),
            };
        } else {
            empty_messages += 1;
        }
    }
}

fn extract_values(queue: &mut Vec<u8>) -> [i32; PATTERN_COUNT] {
    let mut pattern_index: [usize; PATTERN_COUNT] = [0; PATTERN_COUNT];
    let mut value: [i32; PATTERN_COUNT] = [0; PATTERN_COUNT];
    for queue_item in queue.iter() {
        for i in 0..PATTERN_COUNT {
            if pattern_index[i] >= 23 {
            } else if pattern_index[i] >= 19 {
                let e = *queue_item as i32;
                value[i] += e << ((22 - pattern_index[i]) * 8);
                pattern_index[i] += 1;
            } else if *queue_item == PATTERN[i][pattern_index[i]] {
                pattern_index[i] += 1;
            } else {
                pattern_index[i] = 0;
            }
        }
    }
    queue.clear();
    value
}
